﻿using System;
using System.Linq;
using AITB.Solvers;
using AITB.Core;
using AITB.Helpers;
using OPT.Settings;
using AITB.Settings;
using System.Windows.Forms;
using MathNet.Numerics;

namespace OPT.Solvers.Fly
{
    internal class DragonFlyAlgorithmV2 : GenericSolver
    {
        private Problem _problem;
        private Chromosome[] _population;
        private Chromosome[] _deltaPopulation;
        private Chromosome _bestSolution;

        //variant based on pseudocode
        public DragonFlyAlgorithmV2() : base()
        { }

        public override string GetControlParameters()
        {
            throw new NotImplementedException();
        }

        public override void Solve(Problem problem)
        {
            _problem = problem;

            double[] radius = new double[_problem.TotalDimensionality];
            double[] delta_max = new double[_problem.TotalDimensionality];
            double food_fitness = double.MinValue;
            double enemy_fitness = double.MaxValue;
            int _positionFood = 0;
            int _positionEnemy = 0;

            //Chromosome enemy;

            for (int i = 0; i < _problem.TotalDimensionality; i++)
            {
                radius[i] = delta_max[i] = (_problem.Parameters.EncodingInfo.MaxValues[i] - _problem.Parameters.EncodingInfo.MinValues[i]) / 10;
            }


            _population = Chromosome.InitializePopulation(_problem);
            _deltaPopulation = Chromosome.InitializePopulation(_problem);
            _bestSolution = (Chromosome)_population.GetBestSolution().Clone();

            for (int i = 0; i < OptimizerSettings.Generations; i++)
            {
                for (int j = 0; j < _problem.TotalDimensionality; j++)
                {
                    var dif = _problem.Parameters.EncodingInfo.MaxValues[j] - _problem.Parameters.EncodingInfo.MinValues[j];
                    radius[j] = dif / 4 + dif * (i / OptimizerSettings.Generations * 2);
                }
                double w = 0.9 - i * (0.9 - 0.4) / OptimizerSettings.Generations;
                double my_c = 0.1 - i * (0.1 - 0) / (OptimizerSettings.Generations / 2);
                if (my_c < 0) my_c = 0;

                var s = 2 * ExtendedRandom.GenerateRandomNumber(OptimizerSettings.Rng) * my_c; //separation weight
                var a = 2 * ExtendedRandom.GenerateRandomNumber(OptimizerSettings.Rng) * my_c; //alignement weight
                var c = 2 * ExtendedRandom.GenerateRandomNumber(OptimizerSettings.Rng) * my_c; //cohesion weight
                var f = 2 * ExtendedRandom.GenerateRandomNumber(OptimizerSettings.Rng); //food attraction weight
                var e = my_c; //énemy distraction weight

                for (int j = 0; j < OptimizerSettings.PopSize; j++)
                {
                    if (_population[j].Fitness > food_fitness)
                    {
                        food_fitness = _population[j].Fitness;
                        _positionFood = j;
                    }

                    if (_population[j].Fitness < enemy_fitness)
                    {
                        enemy_fitness = _population[j].Fitness;
                        _positionEnemy = j;
                    }
                }

                int index;
                for (int j = 0; j < OptimizerSettings.PopSize; j++)
                {
                    //find all the neighbours
                    int[] neighbours = new int[OptimizerSettings.PopSize];
                    index = -1;
                    for (int k = 0; k < OptimizerSettings.PopSize; k++)
                    {
                        bool isNeighbour = true;
                        for (int pp = 0; pp < _problem.TotalDimensionality; pp++)
                        {
                            if (Distance(_population[j].Data[pp], _population[k].Data[pp]) > radius[pp])
                            {
                                isNeighbour = false;
                                break;
                            }
                        }
                        if (isNeighbour)
                        {
                            index++;
                            neighbours[index] = k;
                            
                        }

                    }


                    double[] S = new double[_problem.TotalDimensionality];
                    double[] A = new double[_problem.TotalDimensionality];
                    double[] F = new double[_problem.TotalDimensionality];
                    double[] C = new double[_problem.TotalDimensionality];
                    double[] C_temp = new double[_problem.TotalDimensionality];
                    double[] Enemy = new double[_problem.TotalDimensionality];
                    bool isAtracted = true;

                    if (index > 0)
                    {

                        for (int k = 0; k < index; k++)
                        {
                            for (int pp = 0; pp < _problem.TotalDimensionality; pp++)
                            {
                                //separation
                                S[pp] += -_population[neighbours[k]].Data[pp] - _population[j].Data[pp];

                                //alignement
                                A[pp] += _deltaPopulation[neighbours[k]].Data[pp];

                                //cohesion
                                C_temp[pp] += _population[neighbours[k]].Data[pp];

                                //atraction
                                F[pp] = _population[_positionFood].Data[pp] - _population[j].Data[pp];

                                //distraction
                                Enemy[pp] = _population[_positionEnemy].Data[pp] + _population[j].Data[pp];
                            }
                        }


                        for (int pp = 0; pp < _problem.TotalDimensionality; pp++)
                        {
                            A[pp] = A[pp] / index;
                            C_temp[pp] = C_temp[pp] / index;
                            C[pp] = C_temp[pp] - _population[j].Data[pp];
                        }

                        for (int pp = 0; pp < _problem.TotalDimensionality; pp++)
                        {
                            _deltaPopulation[j].Data[pp] = w * _deltaPopulation[j].Data[pp] + ExtendedRandom.GenerateRandomNumber(OptimizerSettings.Rng) * A[pp]
                                    + ExtendedRandom.GenerateRandomNumber(OptimizerSettings.Rng) * C[pp]
                                    + ExtendedRandom.GenerateRandomNumber(OptimizerSettings.Rng) * S[pp];

                            _population[j].Data[pp] += _deltaPopulation[j].Data[pp];

                        }
                    }
                    else
                    { 
                        var _levy = Levy(OptimizerSettings.Generations);
                        for (int pp = 0; pp < _problem.TotalDimensionality; pp++)
                        {
                            _population[j].Data[pp] += _levy * _population[j].Data[pp];
                            _deltaPopulation[j].Data[pp] = 0;
                        }
                    }

                                      
                    
                    _population[j].VerifyExcedingLimits(_problem.Parameters.EncodingInfo);


                    _population[j].ComputeFitness(ref _problem);
                    if (FesValues.Contains(GeneralSettings.ActualFes))
                    {
                        Solution = (Chromosome)_population[_population.GetIndexOfBestSolution()].Clone();
                        Solution.WriteSolutionToFile(GeneralSettings.ActualFes);
                    }

                }


                //upgrade the best solution and save it
                index = _population.GetIndexOfBestSolution();
                _population[index].ImproveEntity(problem);
                Solution = (Chromosome)_population[index].Clone();
                if (Solution.Fitness > _bestSolution.Fitness)
                {
                    Solution.WriteSolutionToFile();
                    _bestSolution = (Chromosome)Solution.Clone();
                }
                else
                {
                    _bestSolution.WriteSolutionToFile();
                }

               

                if (OptimizerSettings.PerformanceMetrics)
                    _population.DeterminePerformanceMetrics(i, ref CurrentPopulationDiversity);

                Displayer.AppendText("" + (OptimizerSettings.Generations - i) + " ");
                Application.DoEvents();

                if (i > OptimizerSettings.Generations || GeneralSettings.StopTraining ||
                    Solution.Fitness == 1e+10 || GeneralSettings.ActualFes > OptimizerSettings.MaximFes)
                    i = OptimizerSettings.Generations;

            }

            Solution = _bestSolution;
        }


        private double Distance(double a, double b)
        {
            return Math.Abs(a - b);
        }

        private double Levy(int d)
        {
            var beta = 3 / 2;
            var sigma = Math.Pow(SpecialFunctions.Gamma(1 + beta) * Math.Sin(Math.PI * beta / 2) / SpecialFunctions.Gamma((1 + beta) / 2) * beta * Math.Pow(2, (beta - 1) / 2.0), (1.0 / beta));
            var u = ExtendedRandom.GenerateRandomNumber(MyEnums.RandomNumberGeneratorApproach.ClassicVS, 1, d) * sigma;
            var v = ExtendedRandom.GenerateRandomNumber(MyEnums.RandomNumberGeneratorApproach.ClassicVS, 1, d);
            var step = u / Math.Pow(Math.Abs(v), 1.0 / beta);
            return 0.01 * step;
        }
    }
}
