﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using AITB.Solvers;
using AITB.Core;
using AITB.Helpers;
using OPT.Settings;
using AITB.Settings;
using System.Windows.Forms;

namespace OPT.Solvers.Plants

{
    public class InvasiveWeedOptimization: GenericSolver
    {
        private int _actualFES;
        private Problem _problem;
        private List<Chromosome> _population;

        private int smin = 0;//minimum number of seeds
        private int smax = 5; //maximum number of seeds
        private double exponent = 2; //variation reduction exponent
        private double sigma_initial = 0.5; //initial value of standard deviation
        private double sigma_final = 0.001; //final value of standard deviation

        public InvasiveWeedOptimization() : base()
        { }

        public override void Solve(Problem problem)
        {
            _problem = problem;
            _population = Chromosome.InitializePopulation(_problem,true);


            
            for (int i = 0; i < OptimizerSettings.Generations; i++)
            {
                //update standard deviation
                var sigma = Math.Pow((OptimizerSettings.Generations - i) / (OptimizerSettings.Generations - 1), exponent) * (sigma_initial - sigma_final) + sigma_final;
                
                //determine the cost
                var worstFitnes = _population.Min<Chromosome>(x => x.Fitness);
                var bestFitnes = _population.Max<Chromosome>(x => x.Fitness);

                for (int j = 0; j < OptimizerSettings.PopSize; j++)
                {
                    //formula modified for maximimization
                    var ratio = (_population[j].Fitness - bestFitnes) / (worstFitnes - bestFitnes);
                    //if (ratio < 1E-1) ratio = 1;
                    var S = Math.Floor(smin + (smax - smin) * ratio);

                    for (int k = 0; k < S; k++)
                    {
                        Chromosome newSol = new Chromosome(_problem.Parameters);
                        var index = 0;
                        foreach (var p in _population[j].Data)
                        { 
                            newSol.Data[index] = p + sigma * ExtendedRandom.GenerateRandomNumber(OptimizerSettings.Rng);
                            index++;
                        }
                        newSol.VerifyExcedingLimits(_problem.Parameters.EncodingInfo);
                        newSol.ComputeFitness(ref _problem);

                        _population.Add(newSol);

                        if (FesValues.Contains(GeneralSettings.ActualFes))
                        {
                            Solution = (Chromosome)_population[_population.ToArray().GetIndexOfBestSolution()].Clone();
                            Solution.WriteSolutionToFile(GeneralSettings.ActualFes);
                        }
                    }
                }

                _population.OrderByDescending(x => x.Fitness);
                _population[0].ImproveEntity(_problem);

                var totalLengthPopulation = _population.Count();
                _population.RemoveRange(OptimizerSettings.PopSize-1,totalLengthPopulation-OptimizerSettings.PopSize);

                Solution = (Chromosome)_population[0].Clone();

                Solution.WriteSolutionToFile();

                if (OptimizerSettings.PerformanceMetrics)
                    _population.ToArray().DeterminePerformanceMetrics(i, ref CurrentPopulationDiversity, true);

                Displayer.AppendText("" + (OptimizerSettings.Generations - i) + " ");
                Application.DoEvents();

                if (i > OptimizerSettings.Generations || GeneralSettings.StopTraining ||
                    Solution.Fitness == 1e+10 || GeneralSettings.ActualFes > OptimizerSettings.MaximFes)
                    i = OptimizerSettings.Generations;

            }

        }


        public override string GetControlParameters()
        {
            throw new NotImplementedException();
        }
    }
}
