﻿using AITB.Core;
using AITB.Helpers;
using OPT.Settings;
using System;
using System.Text;

namespace AITB.Solvers.BFO
{
    public class Bacterium : CoreEntity
    {
        //Public swarmCost As Double ' cost from intra-baceria proximity
        //public double prevSwarmCost
        // accumulated measure per swim
        public double health;

        public double prevCost;

        public Bacterium(AITBParameters aitbParameters)
        {
            // random position bacterium
            Data = new double[aitbParameters.EncodingInfo.NoGenes];
            for (var p = 0; p <= aitbParameters.EncodingInfo.NoGenes - 1; p++)
                Data[p] = (aitbParameters.EncodingInfo.MaxValues[p] - aitbParameters.EncodingInfo.MinValues[p]) *
                          ExtendedRandom.GenerateRandomNumber(OptimizerSettings.Rng) +
                          aitbParameters.EncodingInfo.MinValues[p];
            // costs are due to environment and must be computed extermally
            health = 0.0;
        }

        public Bacterium(AITBParameters aitbParameters, double rngParameter)
        {
            // random position bacterium
            Data = new double[aitbParameters.EncodingInfo.NoGenes];
            for (var p = 0; p <= aitbParameters.EncodingInfo.NoGenes - 1; p++)
                Data[p] = (aitbParameters.EncodingInfo.MaxValues[p] - aitbParameters.EncodingInfo.MinValues[p]) *
                          ExtendedRandom.GenerateRandomNumber(OptimizerSettings.Rng, rngParameter) +
                          aitbParameters.EncodingInfo.MinValues[p];
            // costs are due to environment and must be computed extermally
            health = 0.0;
        }

        public Bacterium(ref Bacterium c)
        {
            _isNNChromosome = c._isNNChromosome;
            prevCost = c.prevCost;
            health = c.health;


            Fitness = c.Fitness;
            MSEtesting = c.MSEtesting;
            MSEtraining = c.MSEtraining;
            FitnessInfo = c.FitnessInfo;

            // real-valued
            if (c.Data == null)
            {
                Data = null;
            }
            else
            {
                Data = new double[c.Data.Length];
                Buffer.BlockCopy(c.Data, 0, Data, 0, c.Data.Length * sizeof(double));
            }
        }

        public override string ToString()
        {
            var s = "[ ";
            var sb = new StringBuilder();
            for (var p = 0; p <= Data.Length - 1; p++)
            {
                sb.Append(Data[p].ToString("F2")).Append(" ");
                s += sb.ToString();
            }
            s += "] ";

            s += "cost = " + Fitness.ToString("F2") + " prevCost = " + prevCost.ToString("F2");
            s += " health = " + health.ToString("F2");

            return s;
        }

        public override object Clone()
        {
            var c = this;
            return new Bacterium(ref c);
        }


    }
}